94057c3d3e
PR Test (NPU) / check-changes (push) Has been cancelled
PR Test (NPU) / pr-gate (push) Has been cancelled
PR Test (NPU) / set-image-config (push) Has been cancelled
PR Test (NPU) / stage-b-test-1-npu-a2 (0) (push) Has been cancelled
PR Test (NPU) / stage-b-test-1-npu-a2 (1) (push) Has been cancelled
PR Test (NPU) / stage-b-test-2-npu-a2 (0) (push) Has been cancelled
PR Test (NPU) / stage-b-test-2-npu-a2 (1) (push) Has been cancelled
PR Test (NPU) / stage-b-test-4-npu-a3 (push) Has been cancelled
PR Test (NPU) / stage-b-test-16-npu-a3 (push) Has been cancelled
PR Test (NPU) / multimodal-gen-test-1-npu-a3 (push) Has been cancelled
PR Test (NPU) / multimodal-gen-test-2-npu-a3 (push) Has been cancelled
PR Test (Arm64) / pr-gate (push) Has been cancelled
PR Test (Arm64) / check-changes (push) Has been cancelled
PR Test (Arm64) / build-test (push) Has been cancelled
PR Test (sgl-router) / gate (push) Has been cancelled
PR Test (sgl-router) / tier-1 — lint (push) Has been cancelled
PR Test (sgl-router) / tier-2 — build + test (push) Has been cancelled
PR Test (sgl-router) / tier-3 — docker (placeholder) (push) Has been cancelled
PR Test (sgl-router) / tier-3 — k8s integration (push) Has been cancelled
PR Test (sgl-router) / tier-3 — e2e (push) Has been cancelled
PR Test (sgl-router) / finish (push) Has been cancelled
PR Test (NPU) / single-node-poc (map[name:qwen3_6_27b_w8a8_1p_in64k_out1k_50ms runner:linux-aarch64-a3-2 test_case:test/registered/ascend/performance/qwen3_6_27b/test_npu_qwen3_6_27b_w8a8_1p_in64k_out1k_50ms.py test_type:perf]) (push) Has been cancelled
PR Test (NPU) / pr-test-npu-finish (push) Has been cancelled
PR Test (Xeon) / pr-gate (push) Has been cancelled
PR Test (Xeon) / check-changes (push) Has been cancelled
PR Test (Xeon) / build-test (, xeon-gnr, base-b-test-cpu) (push) Has been cancelled
PR Test (XPU) / check-changes (push) Has been cancelled
PR Test (XPU) / pr-gate (push) Has been cancelled
PR Test (XPU) / stage-a-test-1-gpu-xpu (push) Has been cancelled
PR Test (XPU) / wait-for-stage-a (push) Has been cancelled
PR Test (XPU) / stage-b-test-1-gpu-xpu (push) Has been cancelled
PR Test (XPU) / finish (push) Has been cancelled
CI Model Inventory / build-inventory (push) Has been cancelled
Lint / lint (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark Compilation Check (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark - Manual Policy (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark - Request Processing (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark Summary (push) Has been cancelled
PR Test (SMG) / build-wheel (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on windows (x86_64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on macos (x86_64 - auto) (push) Has been cancelled
PR Test (SMG) / python-unit-tests (push) Has been cancelled
PR Test (SMG) / unit-tests (push) Has been cancelled
PR Test (SMG) / benchmarks (push) Has been cancelled
PR Test (SMG) / chat-completions (push) Has been cancelled
PR Test (SMG) / chat-completions-4gpu (push) Has been cancelled
PR Test (SMG) / e2e (push) Has been cancelled
PR Test (SMG) / docker-build-test (push) Has been cancelled
PR Test (SMG) / k8s-integration (push) Has been cancelled
PR Test (SMG) / finish (push) Has been cancelled
PR Test (SMG) / summarize-benchmarks (push) Has been cancelled
Release SGLang Model Gateway Docker Image / publish (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on macos (aarch64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (aarch64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (x86_64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (aarch64 - musllinux_1_1) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (x86_64 - musllinux_1_1) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / Build SDist (push) Has been cancelled
Release SGLang Model Gateway to PyPI / Upload to PyPI (push) Has been cancelled
Release SGLang Kernels / build-cu129-matrix (aarch64, 12.9, 3.10, arm-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / build-cu129-matrix (x86_64, 12.9, 3.10, x64-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / release-cu129 (push) Has been cancelled
Release SGLang Kernels / build-cu130-matrix (aarch64, 13.0, 3.10, arm-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / build-cu130-matrix (x86_64, 13.0, 3.10, x64-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / release-cu130 (push) Has been cancelled
Release SGLang Kernels / build-rocm-matrix (3.10, 700) (push) Has been cancelled
Release SGLang Kernels / build-rocm-matrix (3.10, 720) (push) Has been cancelled
Release SGLang Kernels / release-rocm700 (push) Has been cancelled
Release SGLang Kernels / release-rocm720 (push) Has been cancelled
Release SGLang Kernels / build-musa43 (43, 3.10) (push) Has been cancelled
Release SGLang Kernels / release-musa43 (push) Has been cancelled
295 lines
8.3 KiB
Python
295 lines
8.3 KiB
Python
# Adapt from https://github.com/fla-org/flash-linear-attention/blob/main/fla/ops/utils/cumsum.py
|
|
# -*- coding: utf-8 -*-
|
|
# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang
|
|
|
|
from typing import Optional
|
|
|
|
import torch
|
|
import triton
|
|
import triton.language as tl
|
|
|
|
from sglang.srt.layers.attention.fla.index import prepare_chunk_indices
|
|
from sglang.srt.layers.attention.fla.utils import check_shared_mem, input_guard
|
|
|
|
BS_LIST = [32, 64] if check_shared_mem() else [16, 32]
|
|
|
|
|
|
# @triton.autotune(
|
|
# configs=[triton.Config({}, num_warps=num_warps) for num_warps in [1, 2, 4, 8]],
|
|
# key=["B", "H", "BT", "IS_VARLEN", "REVERSE"],
|
|
# )
|
|
@triton.jit(do_not_specialize=["T"])
|
|
def chunk_local_cumsum_scalar_kernel(
|
|
s,
|
|
o,
|
|
scale,
|
|
cu_seqlens,
|
|
chunk_indices,
|
|
T,
|
|
B: tl.constexpr,
|
|
H: tl.constexpr,
|
|
BT: tl.constexpr,
|
|
REVERSE: tl.constexpr,
|
|
HAS_SCALE: tl.constexpr,
|
|
IS_VARLEN: tl.constexpr,
|
|
HEAD_FIRST: tl.constexpr,
|
|
):
|
|
i_t, i_bh = tl.program_id(0), tl.program_id(1)
|
|
i_b, i_h = i_bh // H, i_bh % H
|
|
if IS_VARLEN:
|
|
i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(
|
|
chunk_indices + i_t * 2 + 1
|
|
).to(tl.int32)
|
|
bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(
|
|
cu_seqlens + i_n + 1
|
|
).to(tl.int32)
|
|
T = eos - bos
|
|
else:
|
|
bos, eos = i_b * T, i_b * T + T
|
|
|
|
if HEAD_FIRST:
|
|
p_s = tl.make_block_ptr(
|
|
s + bos * H + i_h * T, (T,), (1,), (i_t * BT,), (BT,), (0,)
|
|
)
|
|
p_o = tl.make_block_ptr(
|
|
o + bos * H + i_h * T, (T,), (1,), (i_t * BT,), (BT,), (0,)
|
|
)
|
|
else:
|
|
p_s = tl.make_block_ptr(s + bos * H + i_h, (T,), (H,), (i_t * BT,), (BT,), (0,))
|
|
p_o = tl.make_block_ptr(o + bos * H + i_h, (T,), (H,), (i_t * BT,), (BT,), (0,))
|
|
# [BT]
|
|
b_s = tl.load(p_s, boundary_check=(0,)).to(tl.float32)
|
|
b_o = tl.cumsum(b_s, axis=0)
|
|
if REVERSE:
|
|
b_z = tl.sum(b_s, axis=0)
|
|
b_o = -b_o + b_z[None] + b_s
|
|
if HAS_SCALE:
|
|
b_o *= scale
|
|
tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0,))
|
|
|
|
|
|
@triton.autotune(
|
|
configs=[
|
|
triton.Config({"BS": BS}, num_warps=num_warps, num_stages=num_stages)
|
|
for BS in BS_LIST
|
|
for num_warps in [2, 4, 8]
|
|
for num_stages in [2, 3, 4]
|
|
],
|
|
key=["B", "H", "S", "BT", "IS_VARLEN", "REVERSE", "HAS_SCALE"],
|
|
)
|
|
@triton.jit(do_not_specialize=["T"])
|
|
def chunk_local_cumsum_vector_kernel(
|
|
s,
|
|
o,
|
|
scale,
|
|
cu_seqlens,
|
|
chunk_indices,
|
|
T,
|
|
B: tl.constexpr,
|
|
H: tl.constexpr,
|
|
S: tl.constexpr,
|
|
BT: tl.constexpr,
|
|
BS: tl.constexpr,
|
|
REVERSE: tl.constexpr,
|
|
HAS_SCALE: tl.constexpr,
|
|
IS_VARLEN: tl.constexpr,
|
|
HEAD_FIRST: tl.constexpr,
|
|
):
|
|
i_s, i_t, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
|
|
i_b, i_h = i_bh // H, i_bh % H
|
|
if IS_VARLEN:
|
|
i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(
|
|
chunk_indices + i_t * 2 + 1
|
|
).to(tl.int32)
|
|
bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(
|
|
cu_seqlens + i_n + 1
|
|
).to(tl.int32)
|
|
T = eos - bos
|
|
else:
|
|
bos, eos = i_b * T, i_b * T + T
|
|
|
|
o_i = tl.arange(0, BT)
|
|
if REVERSE:
|
|
m_s = tl.where(o_i[:, None] <= o_i[None, :], 1.0, 0.0)
|
|
else:
|
|
m_s = tl.where(o_i[:, None] >= o_i[None, :], 1.0, 0.0)
|
|
|
|
if HEAD_FIRST:
|
|
p_s = tl.make_block_ptr(
|
|
s + (bos * H + i_h * T) * S,
|
|
(T, S),
|
|
(S, 1),
|
|
(i_t * BT, i_s * BS),
|
|
(BT, BS),
|
|
(1, 0),
|
|
)
|
|
p_o = tl.make_block_ptr(
|
|
o + (bos * H + i_h * T) * S,
|
|
(T, S),
|
|
(S, 1),
|
|
(i_t * BT, i_s * BS),
|
|
(BT, BS),
|
|
(1, 0),
|
|
)
|
|
else:
|
|
p_s = tl.make_block_ptr(
|
|
s + (bos * H + i_h) * S,
|
|
(T, S),
|
|
(H * S, 1),
|
|
(i_t * BT, i_s * BS),
|
|
(BT, BS),
|
|
(1, 0),
|
|
)
|
|
p_o = tl.make_block_ptr(
|
|
o + (bos * H + i_h) * S,
|
|
(T, S),
|
|
(H * S, 1),
|
|
(i_t * BT, i_s * BS),
|
|
(BT, BS),
|
|
(1, 0),
|
|
)
|
|
# [BT, BS]
|
|
b_s = tl.load(p_s, boundary_check=(0, 1)).to(tl.float32)
|
|
b_o = tl.dot(m_s, b_s, allow_tf32=False)
|
|
if HAS_SCALE:
|
|
b_o *= scale
|
|
tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0, 1))
|
|
|
|
|
|
def chunk_local_cumsum_scalar(
|
|
g: torch.Tensor,
|
|
chunk_size: int,
|
|
reverse: bool = False,
|
|
scale: float = None,
|
|
cu_seqlens: Optional[torch.Tensor] = None,
|
|
head_first: bool = False,
|
|
output_dtype: Optional[torch.dtype] = torch.float,
|
|
chunk_indices: Optional[torch.LongTensor] = None,
|
|
) -> torch.Tensor:
|
|
if head_first:
|
|
B, H, T = g.shape
|
|
else:
|
|
B, T, H = g.shape
|
|
assert chunk_size == 2 ** (
|
|
chunk_size.bit_length() - 1
|
|
), "chunk_size must be a power of 2"
|
|
BT = chunk_size
|
|
if chunk_indices is None and cu_seqlens is not None:
|
|
chunk_indices = prepare_chunk_indices(cu_seqlens, BT)
|
|
NT = triton.cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
|
|
g_org, g = g, torch.empty_like(g, dtype=output_dtype or g.dtype)
|
|
grid = (NT, B * H)
|
|
chunk_local_cumsum_scalar_kernel[grid](
|
|
s=g_org,
|
|
o=g,
|
|
scale=scale,
|
|
cu_seqlens=cu_seqlens,
|
|
chunk_indices=chunk_indices,
|
|
T=T,
|
|
B=B,
|
|
H=H,
|
|
BT=BT,
|
|
HEAD_FIRST=head_first,
|
|
REVERSE=reverse,
|
|
HAS_SCALE=scale is not None,
|
|
IS_VARLEN=cu_seqlens is not None,
|
|
num_warps=8,
|
|
num_stages=3,
|
|
)
|
|
return g
|
|
|
|
|
|
def chunk_local_cumsum_vector(
|
|
g: torch.Tensor,
|
|
chunk_size: int,
|
|
reverse: bool = False,
|
|
scale: float = None,
|
|
cu_seqlens: Optional[torch.Tensor] = None,
|
|
head_first: bool = False,
|
|
output_dtype: Optional[torch.dtype] = torch.float,
|
|
chunk_indices: Optional[torch.LongTensor] = None,
|
|
) -> torch.Tensor:
|
|
if head_first:
|
|
B, H, T, S = g.shape
|
|
else:
|
|
B, T, H, S = g.shape
|
|
BT = chunk_size
|
|
if chunk_indices is None and cu_seqlens is not None:
|
|
chunk_indices = prepare_chunk_indices(cu_seqlens, BT)
|
|
NT = triton.cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
|
|
assert chunk_size == 2 ** (
|
|
chunk_size.bit_length() - 1
|
|
), "chunk_size must be a power of 2"
|
|
|
|
g_org, g = g, torch.empty_like(g, dtype=output_dtype or g.dtype)
|
|
|
|
def grid(meta):
|
|
return (triton.cdiv(meta["S"], meta["BS"]), NT, B * H)
|
|
|
|
# keep cumulative normalizer in fp32
|
|
# this kernel is equivalent to
|
|
# g = g.view(B, H, NT, BT, -1).cumsum(-2).view(B, H, T, -1)
|
|
chunk_local_cumsum_vector_kernel[grid](
|
|
s=g_org,
|
|
o=g,
|
|
scale=scale,
|
|
cu_seqlens=cu_seqlens,
|
|
chunk_indices=chunk_indices,
|
|
T=T,
|
|
B=B,
|
|
H=H,
|
|
S=S,
|
|
BT=BT,
|
|
HEAD_FIRST=head_first,
|
|
REVERSE=reverse,
|
|
HAS_SCALE=scale is not None,
|
|
IS_VARLEN=cu_seqlens is not None,
|
|
)
|
|
return g
|
|
|
|
|
|
@input_guard
|
|
def chunk_local_cumsum(
|
|
g: torch.Tensor,
|
|
chunk_size: int,
|
|
reverse: bool = False,
|
|
scale: float = None,
|
|
cu_seqlens: Optional[torch.Tensor] = None,
|
|
head_first: bool = False,
|
|
output_dtype: Optional[torch.dtype] = torch.float,
|
|
chunk_indices: Optional[torch.LongTensor] = None,
|
|
**kwargs,
|
|
) -> torch.Tensor:
|
|
if cu_seqlens is not None:
|
|
assert (
|
|
g.shape[0] == 1
|
|
), "Only batch size 1 is supported when cu_seqlens are provided"
|
|
if len(g.shape) == 3:
|
|
return chunk_local_cumsum_scalar(
|
|
g=g,
|
|
chunk_size=chunk_size,
|
|
reverse=reverse,
|
|
scale=scale,
|
|
cu_seqlens=cu_seqlens,
|
|
head_first=head_first,
|
|
output_dtype=output_dtype,
|
|
chunk_indices=chunk_indices,
|
|
)
|
|
elif len(g.shape) == 4:
|
|
return chunk_local_cumsum_vector(
|
|
g=g,
|
|
chunk_size=chunk_size,
|
|
reverse=reverse,
|
|
scale=scale,
|
|
cu_seqlens=cu_seqlens,
|
|
head_first=head_first,
|
|
output_dtype=output_dtype,
|
|
chunk_indices=chunk_indices,
|
|
)
|
|
else:
|
|
raise ValueError(
|
|
f"Unsupported input shape {g.shape}, "
|
|
f"which should be (B, T, H, D) if `head_first=False` "
|
|
f"or (B, H, T, D) otherwise"
|
|
)
|